# Copyright (c) 2019, 2020 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
**zabel** openapi helper."""

import yaml

BOILERPLATE = {
    "swagger": "2.0",
    "openapi": "3.0.3",
    "info": {"title": "Zabel", "version": "0.10.0"},
    "paths": {
        "/api/": {
            "get": {
                "description": "get available API versions",
                "consumes": ["application/json",],
                "produces": ["application/json",],
                "schemes": ["https"],
                "tags": ["core"],
                "operationId": "getCoreAPIVersions",
                "responses": {
                    "200": {"description": "OK",},
                    "401": {"description": "Unauthorized"},
                },
            }
        },
        "/api/v1/": {
            "get": {
                "description": "get available resources",
                "consumes": ["application/json",],
                "produces": ["application/json",],
                "schemes": ["https"],
                "tags": ["core_v1"],
                "operationId": "getCoreV1APIResources",
                "responses": {
                    "200": {"description": "OK",},
                    "401": {"description": "Unauthorized"},
                },
            }
        },
    },
}

BASE_GROUP = {
    "description": "get information of a group",
    "consumes": ["application/json",],
    "produces": ["application/json",],
    "schemes": ["https"],
    "tags": ["{group}"],
    "operationId": "get{Group}APIGroup",
    "responses": {
        "200": {"description": "OK",},
        "401": {"description": "Unauthorized"},
    },
}


BASE_RESOURCES = {
    "description": "get available resources",
    "consumes": ["application/json",],
    "produces": ["application/json",],
    "schemes": ["https"],
    "tags": ["{group}_{version}"],
    "operationId": "get{Group}{Version}APIResources",
    "responses": {
        "200": {"description": "OK",},
        "401": {"description": "Unauthorized"},
    },
}

BASE_CREATE_PARAMETERS = [
    {"name": "body", "in": "body", "required": True,},
    {
        "uniqueItems": True,
        "type": "string",
        "description": "When present, indicates that modifications should not be persisted. An invalid or unrecognized dryRun directive will result in an error response and no further processing of the request. Valid values are: - All: all dry run stages will be processed",
        "name": "dryRun",
        "in": "query",
    },
    {
        "uniqueItems": True,
        "type": "string",
        "description": "fieldManager is a name associated with the actor or entity that is making these changes. The value must be less than or 128 characters long, and only contain printable characters, as defined by https://golang.org/pkg/unicode/#IsPrint.",
        "name": "fieldManager",
        "in": "query",
    },
]

BASE_LIST_PARAMETERS = [
    {
        "uniqueItems": True,
        "type": "string",
        "description": "The continue option should be set when retrieving more results from the server. Since this value is server defined, clients may only use the continue value from a previous query result with identical query parameters (except for the value of continue) and the server may reject a continue value it does not recognize. If the specified continue value is no longer valid whether due to expiration (generally five to fifteen minutes) or a configuration change on the server, the server will respond with a 410 ResourceExpired error together with a continue token. If the client needs a consistent list, it must restart their list without the continue field. Otherwise, the client may send another list request with the token received with the 410 error, the server will respond with a list starting from the next key, but from the latest snapshot, which is inconsistent from the previous list results - objects that are created, modified, or deleted after the first list request will be included in the response, as long as their keys are after the \"next key\".\n\nThis field is not supported when watch is true. Clients may start a watch from the last resourceVersion value returned by the server and not miss any modifications.",
        "name": "continue",
        "in": "query",
    },
    {
        "uniqueItems": True,
        "type": "string",
        "description": "A selector to restrict the list of returned objects by their fields. Defaults to everything.",
        "name": "fieldSelector",
        "in": "query",
    },
    {
        "uniqueItems": True,
        "type": "string",
        "description": "A selector to restrict the list of returned objects by their labels. Defaults to everything.",
        "name": "labelSelector",
        "in": "query",
    },
    {
        "uniqueItems": True,
        "type": "integer",
        "description": "limit is a maximum number of responses to return for a list call. If more items exist, the server will set the `continue` field on the list metadata to a value that can be used with the same initial query to retrieve the next set of results. Setting a limit may return fewer than the requested amount of items (up to zero items) in the event all requested objects are filtered out and clients should only use the presence of the continue field to determine whether more results are available. Servers may choose not to support the limit argument and will return all of the available results. If limit is specified and the continue field is empty, clients may assume that no more results are available. This field is not supported if watch is true.\n\nThe server guarantees that the objects returned when using continue will be identical to issuing a single list call without a limit - that is, no objects created, modified, or deleted after the first request is issued will be included in any subsequent continued requests. This is sometimes referred to as a consistent snapshot, and ensures that a client that is using limit to receive smaller chunks of a very large result can ensure they see all possible objects. If objects are updated during a chunked list the version of the object that was present at the time the first list result was calculated is returned.",
        "name": "limit",
        "in": "query",
    },
    {
        "uniqueItems": True,
        "type": "string",
        "description": "When specified with a watch call, shows changes that occur after that particular version of a resource. Defaults to changes from the beginning of history. When specified for list: - if unset, then the result is returned from remote storage based on quorum-read flag; - if it's 0, then we simply return what we currently have in cache, no guarantee; - if set to non zero, then the result is at least as fresh as given rv.",
        "name": "resourceVersion",
        "in": "query",
    },
    {
        "uniqueItems": True,
        "type": "integer",
        "description": "Timeout for the list/watch call. This limits the duration of the call, regardless of any activity or inactivity.",
        "name": "timeoutSeconds",
        "in": "query",
    },
    {
        "uniqueItems": True,
        "type": "boolean",
        "description": "Watch for changes to the described resources and return them as a stream of add, update, and remove notifications. Specify resourceVersion.",
        "name": "watch",
        "in": "query",
    },
]

VERB_OPERATION = {
    'list': 'get',
    'create': 'post',
    'deletecollection': 'delete',
    'delete': 'delete',
    'get': 'get',
    'patch': 'patch',
    'update': 'put',
}

GENERIC_VERBS = {'list', 'create', 'deletecollection'}
SPECIFIC_VERBS = {'delete', 'get', 'patch', 'update'}

PREFIX = b'/registry/apiregistration.k8s.io/apiservices/'


def emit_generics(resource, context):
    return {
        VERB_OPERATION[verb]: emit_operation(verb, resource, context)
        for verb in GENERIC_VERBS & set(resource['verbs'])
    }


def emit_specifics(resource, context):
    return {
        VERB_OPERATION[verb]: emit_operation(verb, resource, context)
        for verb in SPECIFIC_VERBS & set(resource['verbs'])
    }


def emit_operation(verb, resource, context):
    namespaced = 'Namespaced' if resource.get('namespaced') else ''
    return {
        'description': f'{verb} objects of kind {resource["kind"]}',
        'consumes': ['*/*'],
        'produces': ['application/json'],
        'tags': [context['tag']],
        'operationId': f'{verb}{context["baseoperation"]}{namespaced}{resource["kind"]}',
        'responses': {
            '200': {'description': 'OK'},
            '401': {'description': 'Unauthorized'},
        },
    }


def emit_subresource(root, resource, context):
    res, sub = resource['name'].split('/')
    if resource.get('namespaced'):
        path = f'{root}/namespaces/{{namespace}}/{res}/{{name}}/{sub}'
    else:
        path = f'{root}/{res}/{{name}}/{sub}'
    verbs = set(resource['verbs'])
    if GENERIC_VERBS & verbs:
        raise ValueError('oops, got generic verb for subresource')
    BOILERPLATE['paths'][path] = emit_specifics(resource, context)


def emit_resource(root, resource, context):
    if resource.get('namespaced'):
        path = f'{root}/namespaces/{{namespace}}/{resource["name"]}'
    else:
        path = f'{root}/{resource["name"]}'
    verbs = set(resource['verbs'])
    if GENERIC_VERBS & verbs:
        BOILERPLATE['paths'][path] = emit_generics(resource, context)
    if SPECIFIC_VERBS & verbs:
        BOILERPLATE['paths'][f'{path}/{{name}}'] = emit_specifics(
            resource, context
        )


def ensure_emit_group(groupversion, context):
    group, version = groupversion.split('/')
    key = f'/apis/{group}/'
    if key not in BOILERPLATE['paths']:
        name = group.split('.')[0] if group.endswith('.k8s.io') else group
        information = BASE_GROUP.copy()
        information['tags'] = [name.lower()]
        information['operationId'] = f'get{context["baseoperation"]}ApiGroup'
        BOILERPLATE['paths'][key] = {'get': information}


def emit_group_version(root, groupversion, context):
    group, version = groupversion.split('/')
    name = group.split('.')[0]
    resources = BASE_RESOURCES.copy()
    resources['tags'] = [context['tag']]
    resources['operationId'] = f'get{context["baseoperation"]}APIResources'
    BOILERPLATE['paths'][f'{root}/'] = {'get': resources}


def make_tag(groupversion: str) -> str:
    """Make tag corresponding to groupversion."""
    if '/' in groupversion:
        group, version = groupversion.split('/')
        splits = group.split('.')
        lhs = splits[0].lower() + ''.join(s.title() for s in splits[1:])
        return f'{lhs}_{version.lower()}'
    else:
        return f'core_{groupversion.lower()}'


def make_operationname(groupversion: str) -> str:
    """Make CamelCase base operation name."""
    if '/' in groupversion:
        group, version = groupversion.split('/')
        if group.endswith('.k8s.io'):
            group = group[:-7]
        return ''.join(s.title() for s in group.split('.')) + version.title()
    else:
        return f'Core{groupversion.title()}'


def get_openapi(cluster):
    apis = [yaml.safe_load(r) for r, _ in cluster.etcd.get_prefix(PREFIX)]

    for api in apis:
        groupversion = api['groupVersion']
        context = {
            'tag': make_tag(groupversion),
            'baseoperation': make_operationname(groupversion),
        }
        if '/' in groupversion:
            root = f'/apis/{groupversion}'
            ensure_emit_group(groupversion, context)
            emit_group_version(root, groupversion, context)
        else:
            root = f'/api/{groupversion}'

        for resource in api['resources']:
            if '/' in resource['name']:
                emit_subresource(root, resource, context)
            else:
                emit_resource(root, resource, context)
    return BOILERPLATE



# Copyright (c) 2019 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
The high-level **zabel.fabric** library.

"""



# Copyright (c) 2019, 2020 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
**amac** etcd.

This module implements a naïve etcd service.

The public API it provides is a strict subset of the API offered by
the **python-etcd3** library.  Switching to it should be seamless. (The
opposite is not true, and not all features are implemented.)
"""


from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from datetime import datetime

import os
import pickle
import queue
import threading

########################################################################
########################################################################

# VData: data (bytes) with an attached revision number (int)
# HVData: when a key is deleted, its successive values (list of VData)
#         with an attached revision number (int)
# KVData: the 'life' of a key, with the list of changes of its current
#         incarnation (list of VData), as well as a list of its previous
#         incarnations (list of HVData).

# Finding the value of a key at a given time (revision) is as follows:
#
# If the revision is equal or higher than the revision of the initial
# current incarnation value, the value is the value of the change that
# has a revision that is lesser or equal to the desired revision.
#
# If the revision predates the revision of the initial current
# incarnation value, iterate over the previous incarnations, in
# decreasing order.  If the revision is higher or equal than the
# revision attached to the previous incarnation, the key had no value
# at the revision.  Otherwise proceed as in the current incarnation
# step.
#
# If no value is found, the key had no value at this revision.

VData = Tuple[bytes, int]  # data, revision
HVData = Tuple[List[VData], int]  # changes, revision
KVData = Tuple[List[VData], List[HVData]]  # changes, history

# Helpers

VALUE = 0
REVISION = 1


def _current(kvd: KVData, field: int) -> Any:
    return kvd[0][-1][field]


def _initial(kvd: KVData, field: int) -> Any:
    return kvd[0][0][field]


def _increment_last_byte(byte_string: bytes) -> bytes:
    array = bytearray(byte_string)
    array[-1] = array[-1] + 1
    return bytes(array)


def _to_bytes(maybe_bytestring: Union[str, bytes]) -> bytes:
    """Encode string to bytes.

    Convenience function to do a simple encode('utf-8') if the input is not
    already bytes. Returns the data unmodified if the input is bytes.
    """
    if isinstance(maybe_bytestring, bytes):
        return maybe_bytestring
    return maybe_bytestring.encode('utf-8')


# Events classes and helpers


class Event:
    """An event class for watchers.

    Do not use this class directly, use one of its subclass, #PutEvent
    and #DeleteEvent.
    """

    def __init__(
        self, key: bytes, value: Optional[bytes], prev_value: Optional[bytes]
    ) -> None:
        self.key = key
        self.value = value
        self.prev_value = prev_value

    def __str__(self) -> str:
        return f'{self.__class__.__name__}: {self.key!r}, {self.value!r}'

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.key!r}, {self.value!r}>'


class PutEvent(Event):
    """Key creation or modification event."""


class DeleteEvent(Event):
    """Key deletion or expiration event."""


def _events_notifier(
    event_queue: queue.Queue,
    watchers: List[Tuple[bytes, bytes, Callable[..., None]]],
) -> None:
    while True:
        event = event_queue.get()
        for watcher in watchers:
            if not watcher:
                continue
            range_start, range_end, callback = watcher
            key = _to_bytes(event.key)
            if range_end is None:
                if key == range_start:
                    callback(event)
            elif range_start <= key < range_end:
                callback(event)


# Metadata class


class KVMetadata:
    """A container for key metadata."""

    def __init__(self, key: bytes, kv: KVData) -> None:
        self.key = key
        self.create_revision = _initial(kv, REVISION)
        self.mod_revision = _current(kv, REVISION)
        self.version = len(kv[0])
        self.lease_id = None

    def __str__(self) -> str:
        return (
            f'{self.__class__.__name__}: {self.key!r}, version={self.version}'
        )

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.key!r}, version={self.version}>'


# EtcdClient class

WatcherEntry = Tuple[bytes, Optional[bytes], Callable[..., None]]


class EtcdClient:
    """An etcd-compatible implementation.

    The public API it provides is a strict subset of the API offered by
    the **python3-etcd** library.  Switching to it should be seamless.
    (The opposite is not true, as not all features are implemented.)

    `sort_order` is one of `'ascend`', `'descend'` or None.

    `sort_target` is one of `'key'`, `'version'`, `'create'`, `'mod'``,
    or `'value'.

    Instances of this class can act as context managers:

    ```python
    with EtcdClient(...) as etcd:
        ...
    ```
    """

    store: Dict[bytes, KVData]
    revisions: Dict[int, datetime]
    revision: int

    def __init__(self, host: str = 'localhost', port: int = 2379):
        """..."""
        self._filename = f'{host}_{port}.pkl'
        if os.path.isfile(self._filename):
            with open(self._filename, 'rb') as f:
                self.revision, self.store, self.revisions = pickle.load(f)
        else:
            self.revision, self.store, self.revisions = 1, {}, {}
        self.watchers: List[Optional[WatcherEntry]] = []
        self.event_queue = queue.Queue()
        self.event_thread = threading.Thread(
            target=_events_notifier,
            args=[self.event_queue, self.watchers],
            daemon=True,
        )
        self.event_thread.start()

    def __str__(self) -> str:
        return f'{self.__class__.__name__}: {self._filename}'

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self._filename!r}>'

    ## Context manager helpers

    def __enter__(self) -> 'EtcdClient':
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()

    ## Revision helpers

    def _new_revision(self) -> int:
        self.revision += 1
        self.revisions[self.revision] = datetime.now()
        return self.revision

    ## Public API

    def close(self) -> None:
        """Snapshot and close the database."""
        self.snapshot()

    def snapshot(self, filename: Optional[str] = None) -> None:
        """Snapshot the database.

        # Optional parameters

        - filename: a non-empty string or None (None by default)
        """
        with open(filename or self._filename, 'wb') as f:
            pickle.dump(
                (self.revision, self.store, self.revisions),
                f,
                pickle.HIGHEST_PROTOCOL,
            )

    def get(self, key: bytes) -> Tuple[Optional[bytes], Optional[KVMetadata]]:
        """Get the value of a key.

        # Required parameters

        - key: a non-empty bytes string

        # Returned value

        A (value, metadata) tuple.  If `key` is not present, returns
        (None, None).
        """
        key = _to_bytes(key)

        if key not in self.store or not self.store[key][0]:
            return None, None
        kvb = self.store[key]
        return _current(kvb, VALUE), KVMetadata(key, kvb)

    def get_prefix(
        self,
        key_prefix: bytes,
        sort_order: Optional[str] = None,
        sort_target: str = 'key',
    ) -> Iterator[Tuple[bytes, KVMetadata]]:
        """Get a range of keys with a prefix.

        # Required parameters

        - key_prefix: a non-empty bytes string

        # Optional parameters

        - sort_target: a non-empty string or None (None by default)
        - sort_target: a non-empty string (`'key'` by default)

        # Returned value

        A sequence of (value, metadata) tuples.
        """
        key_prefix = _to_bytes(key_prefix)
        return self.get_range(
            key_prefix,
            _increment_last_byte(key_prefix),
            sort_order,
            sort_target,
        )

    def get_range(
        self,
        range_start: bytes,
        range_end: bytes,
        sort_order: Optional[str] = None,
        sort_target: str = 'key',
    ) -> Iterator[Tuple[bytes, KVMetadata]]:
        """Get a range of keys.

        # Required parameters

        - range_start: a non-empty bytes string
        - range_end: a non-empty bytes string

        # Returned value

        A sequence of (value, metadata) tuples.
        """
        keys = [
            k
            for k in self.store
            if range_start <= k < range_end and self.store[k][0]
        ]
        for k in keys:
            kvb = self.store[k]
            yield _current(kvb, VALUE), KVMetadata(k, kvb)

    def put(
        self, key: bytes, value: bytes, lease: int = 0, prev_kv: bool = False
    ) -> Optional[bytes]:
        """Save a value.

        # Required parameters

        - key: a non-empty bytes string
        - value: a bytes string

        # Optional parameters

        - lease: an integer (0 by default)
        - prev_kv: a boolean (False by default)
        """
        key = _to_bytes(key)
        value = _to_bytes(value)
        pair = (value, self._new_revision())

        kvb = self.store.get(key, ([], []))
        prev_value = _current(kvb, VALUE) if kvb[0] else None
        kvb[0].append(pair)
        self.store[key] = kvb
        event = PutEvent(key, value, prev_value)
        self.event_queue.put(event)
        return prev_value if prev_kv else None

    def delete(self, key: bytes, prev_kv: bool = False) -> bool:
        """Delete a single key.

        # Required parameters

        - key: a non-empty bytes string

        # Optional parameters

        - prev_kv: a boolean (False by default)

        # Returned values

        A boolean.  True if the deletion was successful, False
        otherwise.
        """
        key = _to_bytes(key)

        if key not in self.store or not self.store[key][0]:
            return False
        kvb = self.store[key]
        pair = (kvb[0], self._new_revision())
        self.store[key] = ([], kvb[1] + [pair])
        event = DeleteEvent(key, None, _current(kvb, VALUE))
        self.event_queue.put(event)
        return True

    def delete_prefix(self, prefix: bytes) -> int:
        """Delete a range of keys with a prefix.

        The operation is atomic, in the sense that all deleted keys are
        deleted at the same revision.

        # Required parameters

        - prefix: a non-empty bytes string

        # Returned value

        An integer, the number of deleted keys.
        """
        prefix = _to_bytes(prefix)

        keys = [
            k for k in self.store if k.startswith(prefix) and self.store[k][0]
        ]
        if not keys:
            return 0
        _revision = self._new_revision()
        values = []
        for k in keys:
            kvb = self.store[k]
            values.append(_current(kvb, VALUE))
            self.store[k] = ([], kvb[1] + [(kvb[0], _revision)])
        for k, value in zip(keys, values):
            self.event_queue.put(DeleteEvent(k, None, value))
        return len(keys)

    def replace(
        self, key: bytes, initial_value: bytes, new_value: bytes
    ) -> bool:
        """Atomically replace the value of a key with a new value.

        # Required parameters

        - key: a non-empty bytes string
        - initial_value: a bytes string
        - new_value: a bytes string

        # Returned value

        A boolean.  True if the replace operation was successful, False
        otherwise.
        """
        if self.get(key)[0] == initial_value:
            self.put(key, new_value)
            return True
        return False

    def add_watch_callback(
        self, key: bytes, callback: Callable[..., None], **kwargs: Any
    ) -> int:
        """Watch a key or range of keys and call a callback on every event.

        # Required parameters

        - key: a non-empty bytes string
        - callback: a function

        # Returned value

        An integer.  It can be used to cancel the watch.  Refer to
        #cancel_watch() for more information.
        """
        self.watchers.append(
            (_to_bytes(key), kwargs.get('range_end', None), callback)
        )
        return len(self.watchers)

    def add_watch_prefix_callback(
        self, key_prefix: bytes, callback: Callable[..., None], **kwargs: Any
    ) -> int:
        """Watch a prefix and call a callback on every event.

        # Required parameters

        - key_prefix: a non-empty bytes string
        - callback: a function

        # Returned value

        An integer.  It can be used to cancel the watch.  Refer to
        #cancel_watch() for more information.
        """
        kwargs['range_end'] = _increment_last_byte(_to_bytes(key_prefix))

        return self.add_watch_callback(key_prefix, callback, **kwargs)

    def cancel_watch(self, watch_id: int) -> None:
        """Stop watching a key or range of keys.

        # Required parameters

        - watch_id: an integer

        # Returned value

        None.
        """
        self.watchers[watch_id] = None

    def compact(self, revision: int) -> None:
        """Compact the event history up to a given revision.

        App superseded keys with a revision less than the compaction
        revision will be removed.
        """
        raise NotImplementedError


def client(host: str = 'localhost', port: int = 2379) -> EtcdClient:
    """Create a client."""
    return EtcdClient(host, port)



# Copyright (c) 2019, 2020 Martin Lafaix (martin.lafaix@external.engie.com)
#
# This program and the accompanying materials are made
# available under the terms of the Eclipse Public License 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0/
#
# SPDX-License-Identifier: EPL-2.0

"""
**zabel** clusters.

This module depends on the public **pyyaml** library.

https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.2.md

etcd and clusters:

<https://medium.com/better-programming/a-closer-look-at-etcd-the-brain-of-a-kubernetes-cluster-788c8ea759a5>

etcd keys follows the kubernetes conventions:

/registry/{resource}/{namespace}/{name}

TODO

system:controller:managedservice-controller
system:controller:utility-controller
system:controller:realm-controller

system:controller:service-controller
"""

from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import base64
import datetime
import json
import pkgutil
import re
import uuid
import random
import sys
import threading

import jwt
import yaml

from zabel.commons.interfaces import ApiService
from zabel.commons.servers import entrypoint
from zabel.commons.utils import patch as patchdict

from .openapi import get_openapi
from .etcd3 import client, Event, PutEvent


########################################################################
## Constants

Object = Dict[str, Any]

NAME_PATTERN = r'^[0-9a-zA-Z]+([0-9A-Za-z-_.]*[0-9a-zA-Z])?$'
LABEL_PATTERN = r'^([^/]+/)?([0-9A-Za-z-_.]{1,63})$'
DNS_LABEL_PATTERN = r'^(?![0-9]+$)(?!-)[a-z0-9-]{1,63}(?<!-)$'

KEY = r'[a-z0-9A-Z-_./]+'
VALUE = r'[a-z0-9A-Z-_.]+'
EQUAL_EXPR = rf'^({KEY})\s*([=!]?=)\s*({VALUE})(?:,|$)'
SET_EXPR = rf'^({KEY})\s+(in|notin)\s+\(({VALUE}(\s*,\s*{VALUE})*)\)(?:,|$)'
EXISTS_EXPR = rf'^{KEY}(?:,|$)'
NEXISTS_EXPR = rf'^!{KEY}(?:,|$)'


########################################################################
## Routes

API_ROUTE = '/api'
APIGROUP_ROUTE = '/apis'

APISERVICE_ROUTES = [
    f'{API_ROUTE}/v1',
    f'{APIGROUP_ROUTE}/{{group}}/{{version}}',
]
CLUSTER_ROUTES = [f'{root}/{{kind}}' for root in APISERVICE_ROUTES]
NAMESPACED_ROUTES = [
    f'{root}/namespaces/{{namespace}}/{{kind}}' for root in APISERVICE_ROUTES
]
CREATE_ROUTES = CLUSTER_ROUTES + NAMESPACED_ROUTES
DIRECT_ROUTES = [f'{root}/{{name}}' for root in CREATE_ROUTES]
STATUS_ROUTES = [f'{root}/status' for root in DIRECT_ROUTES]

DEFAULT_HEADERS = {
    'Content-Type': 'application/json',
    'Strict-Transport-Security': 'max-aincludeSubDomains',
    'X-Frame-Options': 'SAMEORIGIN',
    'X-Content-Type-Options': 'nosniff',
    'Referrer-Policy': 'no-referrer',
    'Content-Security-Policy': 'default-src \'none\'',
}


########################################################################
## Keys templates

APISERVICE_PREFIX = b'/registry/apiregistration.k8s.io/apiservices/'
DEFAULT_NAMESPACE_KEY = b'/registry/namespaces/default'
NAMESPACED_KEY_TEMPLATE = '/registry/{resource}/{namespace}/{name}'
NAMESPACED_PREFIX_TEMPLATE = '/registry/{resource}/{namespace}/'
CLUSTER_KEY_TEMPLATE = '/registry/{resource}/{name}'
CLUSTER_PREFIX_TEMPLATE = '/registry/{resource}/'

CRD_PREFIX = b'/registry/customresourcedefinitions/'

MANAGEDSERVICE_KIND = 'managedservices'
MANAGEDSERVICE_PREFIX = b'/registry/managedservices/'
UTILITY_KIND = 'utilities'
UTILITY_PREFIX = b'/registry/utilities'
REALM_KIND = 'realms'
REALM_PREFIX = b'/registry/realms/'
PLATFORM_KIND = 'platforms'
PLATFORM_PREFIX = b'/registry/platforms/'


########################################################################
## Bootstrap

APISERVICE_KEY_TEMPLATE = (
    '/registry/apiregistration.k8s.io/apiservices/{version}.{group}'
)

APISERVICE_TEMPLATE = '''{{
    "kind": "APIResourceList",
    "apiVersion": "v1",
    "groupVersion": "{group}/{version}",
    "resources": []
}}'''

DEFAULT_NAMESPACE_NAME = 'default'
DEFAULT_NAMESPACE = {
    'apiVersion': 'v1',
    'kind': 'Namespace',
    'metadata': {'name': DEFAULT_NAMESPACE_NAME},
}


########################################################################
## Helpers

## API Server


REASON_STATUS = {
    'OK': 200,
    'Created': 201,
    'NoContent': 204,
    'BadRequest': 400,
    'Unauthorized': 401,
    'Forbidden': 403,
    'NotFound': 404,
    'AlreadyExists': 409,
    'Conflict': 409,
    'Invalid': 422,
}


def status(
    reason: str, message: str, details: Optional[Dict[str, Any]] = None
) -> Object:
    """Return status object."""
    code = REASON_STATUS[reason]
    return {
        'kind': 'Status',
        'apiVersion': 'v1',
        'metadata': {},
        'status': 'Success' if code // 100 == 2 else 'Failure',
        'message': message,
        'reason': reason,
        'details': details,
        'code': code,
    }


def make_items(kind: str, what: List[Object]) -> Object:
    """Return list object."""
    return {'apiVersion': 'v1', 'kind': f'{kind}List', 'items': what}


## Validity checking


def _is_dns_label(value: Any) -> bool:
    return (
        isinstance(value, str)
        and re.match(DNS_LABEL_PATTERN, value) is not None
    )


def _is_dns_domain(value: Any) -> bool:
    if not isinstance(value, str) or len(value) > 253:
        return False
    return all(_is_dns_label(segment) for segment in value.split('.'))


def _is_label_key(value: Any) -> bool:
    if not isinstance(value, str):
        return False
    match = re.match(LABEL_PATTERN, value)
    if not match:
        return False
    prefix, name = match.groups()
    if prefix and not _is_dns_domain(prefix[:-1]):
        return False
    return re.match(NAME_PATTERN, name) is not None


def _is_label_value(value: Any) -> bool:
    if not isinstance(value, str):
        return False
    return len(value) <= 63 and re.match(NAME_PATTERN, value) is not None


## Selectors helpers


def _split_exprs(exprs: str) -> List[str]:
    """Split a comma-separated list of expressions.

    # Required parameters

    - exprs: a string

    # Returned value

    A (possibly empty) list of _expressions_.  An expression is a
    string, stripped.
    """
    result = []
    while exprs:
        match = re.match(SET_EXPR, exprs)
        if not match:
            match = re.match(EQUAL_EXPR, exprs)
        if not match:
            match = re.match(EXISTS_EXPR, exprs)
        if not match:
            match = re.match(NEXISTS_EXPR, exprs)
        if not match:
            raise ValueError(f'Invalid expression {exprs}')
        result.append(exprs[: match.end()].strip())
        exprs = exprs[match.end() :].strip()

    return result


def _resolve_path(path: str, obj: Object) -> Tuple[bool, Optional[str]]:
    def inner(items, obj) -> Tuple[bool, Optional[str]]:
        head, rest = items[0], items[1:]
        if head in obj:
            return (True, obj[head]) if not rest else inner(rest, obj[head])
        return False, None

    return inner(path.split('.'), obj)


def _evaluate_fields(req: str, obj: Object) -> bool:
    if req == '':
        return True
    if re.match(EXISTS_EXPR, req):
        return _resolve_path(req, obj)[0]
    if re.match(NEXISTS_EXPR, req):
        return not _resolve_path(req[1:], obj)[0]
    expr = re.match(SET_EXPR, req)
    if expr:
        key, ope, list_, _ = expr.groups()
        found, value = _resolve_path(key, obj)
        if found:
            values = [v.strip() for v in list_.split(',')]
            if ope == 'in':
                return value in values
            return value not in values
        return ope == 'notin'
    expr = re.match(EQUAL_EXPR, req)
    if expr is None:
        raise ValueError(f'Invalid expression {req}.')
    key, ope, expected = expr.groups()
    found, value = _resolve_path(key, obj)
    if found:
        if ope in ('=', '=='):
            return value == expected
        return value != expected
    return ope == '!='


def _evaluate(req: str, labels: Mapping[str, str]) -> bool:
    """Evaluate whether req matches labels.

    # Required parameters

    - req: a string
    - labels: a dictionary

    # Returned value

    A boolean.  True if `req` is satisfied by `labels`, False otherwise.

    # Raised exceptions

    A _ValueError_ exception is raised if `req` is not a valid
    expression.
    """
    if req == '':
        return True
    if re.match(EXISTS_EXPR, req):
        return req in labels
    if re.match(NEXISTS_EXPR, req):
        return req[1:] not in labels
    expr = re.match(SET_EXPR, req)
    if expr:
        key, ope, list_, _ = expr.groups()
        if key in labels:
            values = [v.strip() for v in list_.split(',')]
            if ope == 'in':
                return labels[key] in values
            return labels[key] not in values
        return ope == 'notin'
    expr = re.match(EQUAL_EXPR, req)
    if expr is None:
        raise ValueError(f'Invalid expression {req}.')
    key, ope, value = expr.groups()
    if key in labels:
        if ope in ('=', '=='):
            return labels[key] == value
        return labels[key] != value
    return ope == '!='


def _match_field_selector(obj: Object, selector: str) -> bool:
    """Return True if the object matches the selector."""
    return all(_evaluate_fields(sel, obj) for sel in _split_exprs(selector))


def _match_label_selector(obj: Object, selector: str) -> bool:
    """Return True if the service matches the selector.

    An empty selector always matches.

    The complete selector feature has been implemented.  `selector` is
    of form:

        expr[,expr]*

    where `expr` is one of `key`, `!key`, or `key op value`, with
    `op` being one of `=`, `==`, or `!=`.  The
    `key in (value[, value...])` and `key notin (value[, value...])`
    set-based requirements are also implemented.

    # Required parameters

    - obj: a Definition (a dictionary)
    - selector: a string

    # Returned value

    A boolean.
    """
    return _match_field_selector(obj['metadata'].get('labels', {}), selector)


class Cluster(ApiService):
    """Clusters.

    Clusters are collections of _objects_.

    Each object has a name, a definition and a status.  Most object are
    attached to a _namespace_.

    Namespaces are objects too, but they are attached to a cluster, not
    to another namespace.  In all other aspects, they are objects.

    All objects are stored in an etcd database as JSON strings.
    """

    def __init__(self, host: str = 'localhost', port: int = 8080) -> None:
        """Create a new cluster.

        The cluster will be bootstrapped if applicable.

        A bootstrapped cluster has a default namespace (`'default'`) and
        contains a set of _custom resource definitions_:

        - Realm
        - Domain
        - Platform
        - ManagedService
        - Utility
        - ManagedProject
        - Secret

        Clusters persist their state.

        After initialization, services (realms, utilities and managed
        services) are started and ready.

        # TODO

        Should not start services but instead have a `run()` method that
        starts the cluster and serves its declared APIs.
        """
        self.host = host
        self.port = port
        self.services = {}
        with open('/etc/zabel/zabel-publickey') as key:
            self._publickey = key.read()
        self.start()

    def start(self):
        """Start cluster.

        If the cluster has never been initialized, it will be
        bootstrapped.

        Services, watchers, and APIs are initialized and started.

        This method does not return.
        """
        with client() as self.etcd:
            self.start_apiserver()
            self.start_watchers()
            if self.etcd.get(DEFAULT_NAMESPACE_KEY) == (None, None):
                self.bootstrap()
            self.start_services()

    def bootstrap(self) -> None:
        """Bootstrap cluster.

        The `etcd` service is expected to be up and running.
        """
        for apiservice in yaml.safe_load_all(
            pkgutil.get_data('zabel.fabric', 'standalone/coreapis.yaml')
        ):
            groupversion = apiservice['groupVersion']
            if '/' in groupversion:
                group, version = groupversion.split('/')
            else:
                group, version = '', groupversion
            self.etcd.put(
                APISERVICE_KEY_TEMPLATE.format(group=group, version=version),
                json.dumps(apiservice),
            )
        self.create('namespaces', DEFAULT_NAMESPACE)

    # Controllers

    def start_apiserver(self) -> None:
        """Start APIs.

        The API server is started in a separate thread, running in
        normal mode (i.e., not in daemon mode).
        """

        def wrap(handler, rbac: bool):
            def inner(*args, **kwargs):
                for header, value in DEFAULT_HEADERS.items():
                    self.response.headers[header] = value
                if rbac:
                    try:
                        user = self._ensure_authn()
                        self._ensure_authz(user)
                    except ValueError as err:
                        response = err.args[0]
                        self.response.status = response['code']
                        return response
                self._ensure_admission()
                if self.request.query.limit:
                    kwargs['limit'] = self.request.query.limit
                if self.request.query.labelSelector:
                    kwargs['labelselector'] = self.request.query.labelSelector
                if self.request.query.fieldSelector:
                    kwargs['fieldselector'] = self.request.query.fieldSelector
                if self.request.json:
                    kwargs['body'] = self.request.json
                elif self.request.body:
                    body = self.request.body.read()
                    if body:
                        kwargs['body'] = yaml.safe_load(body)
                try:
                    result = json.dumps(handler(*args, **kwargs))
                    return result
                except ValueError as err:
                    response = err.args[0]
                    self.response.status = response['code']
                    return response

            return inner

        from bottle import route, run, request, response

        for name in dir(self):
            method = getattr(self, name)
            for endpoint in getattr(method, 'entrypoint routes', []):
                route(
                    path=endpoint['path'].replace('{', '<').replace('}', '>'),
                    method=endpoint['methods'],
                    callback=wrap(method, endpoint['rbac']),
                )
        self.request = request
        self.response = response

        self.api_server = threading.Thread(
            target=run, kwargs={'host': self.host, 'port': self.port},
        )
        self.api_server.start()

    def start_watchers(self) -> None:
        """Start watches of interest.

        The base controller watches services (realms, managed services
        and utilities).
        """
        # self.etcd.add_watch_prefix_callback(REALM_PREFIX, self.handle_event)
        # self.etcd.add_watch_prefix_callback(UTILITY_PREFIX, self.handle_event)
        self.etcd.add_watch_prefix_callback(
            b'/registry/deployments/', self.handle_event
        )
        self.etcd.add_watch_prefix_callback(CRD_PREFIX, self.handle_crdevent)

    def start_services(self) -> None:
        """Start services.

        Ensure services are up and public.  Used at cluster startup.
        """
        for namespace in self.list_allnamespaces('namespaces')['items']:
            _ns = namespace['metadata']['name']
            _ns_services = self.list_namespaced('deployments', namespace=_ns)[
                'items'
            ]
            for service in _ns_services:
                name = service['metadata']['name']
                fullname = name + '.' + service['metadata']['namespace']
                if fullname not in self.services:
                    srv = self._make_service(service)
                    self.services[fullname] = srv
            for name, srv in self.services.items():
                if name.endswith(f'.{_ns}'):
                    self._publish(name.split('.')[0], srv, _ns)

    def handle_crdevent(self, event: Event) -> None:
        """Handle watch events for CustomResourceDefinition."""
        if isinstance(event, PutEvent):
            manifest = json.loads(event.value)
            spec, grp = manifest['spec'], manifest['spec']['group']
            for ver in spec['versions']:
                key = APISERVICE_KEY_TEMPLATE.format(
                    group=grp, version=ver['name']
                )
                val, _ = self.etcd.get(key)
                if val is None:
                    val = APISERVICE_TEMPLATE.format(
                        group=grp, version=ver['name']
                    )
                apiservice = json.loads(val)
                for resource in apiservice['resources']:
                    if resource['name'] == spec['names']['plural']:
                        resource['singularName'] = spec['names']['singular']
                        resource['kind'] = spec['names']['kind']
                        # resource['shortNames'] = spec['names']['shortNames']
                        resource['namespaced'] = spec['scope'] == 'Namespaced'
                        break
                else:
                    apiservice['resources'].append(
                        {
                            'name': spec['names']['plural'],
                            'singularName': spec['names']['singular'],
                            'kind': spec['names']['kind'],
                            # 'shortNames': spec['names']['shortNames'],
                            'namespaced': spec['scope'] == 'Namespaced',
                            'verbs': [
                                'create',
                                'delete',
                                'deletecollection',
                                'get',
                                'list',
                                'patch',
                                'update',
                                'watch',
                            ],
                        }
                    )
                self.etcd.put(key, json.dumps(apiservice))
        else:
            ...

    def handle_event(self, event: Event) -> None:
        """Handle watch events.

        If handling a _PutEvent_, add or update existing service and
        propagate to all services.

        If handling a _DeleteEvent_, remove existing service and
        propagate to all services.

        (Services are accessible to all other services as globals.)
        """
        manifest = json.loads(event.value)
        name = manifest['metadata']['name']
        namespace = manifest['metadata']['namespace']
        if isinstance(event, PutEvent):
            srv = self._make_service(manifest)
            self._publish(name, srv, namespace)
        else:
            self._unpublish(name, namespace)

    # Helpers

    def _get_env_value(
        self, definition: Dict[str, Any], namespace: str
    ) -> Optional[Union[str, Dict[str, Any]]]:
        if 'valueFrom' in definition:
            ref = definition['valueFrom']['secretKeyRef']
            secret = self.get('secrets', ref['name'], namespace=namespace)
            return str(base64.b64decode(secret['data'][ref['key']]), 'utf-8')
        return definition['value']

    def _make_service(self, srv: Object) -> Any:
        """Return a new service instance.

        Expects `srv` to be a deployment-like format.

        # Required parameters

        - srv: an object ( dictionary)

        # Returned value

        An instance of a subclass of _Service_.
        """
        _namespace = srv['metadata']['namespace']
        _container = srv['spec']['template']['spec']['containers'][0]
        _modulename, _classname = _container['image'].rsplit('.', 1)
        _module = __import__(_modulename)
        for _name in _modulename.split('.')[1:]:
            _module = getattr(_module, _name)
        _class = getattr(_module, _classname)
        _env = {
            definition['name']: self._get_env_value(definition, _namespace)
            for definition in _container.get('env', [])
        }
        return _class(_container['name'], _env)

    def _publish(self, name: str, srv: Any, namespace: str) -> None:
        """Add global definition in namespace."""
        self.services[f'{name}.{namespace}'] = srv
        public = name.replace('-', '_')
        for service, instance in self.services.items():
            if service.endswith(f'.{namespace}') and not service.startswith(
                f'{name}.'
            ):
                setattr(sys.modules[instance.__module__], public, srv)

    def _unpublish(self, name: str, namespace: str) -> None:
        """Remove global definition in namespace."""
        if name in self.services:
            del self.services[f'{name}.{namespace}']
        public = name.replace('-', '_')
        for service, instance in self.services.items():
            if service.endswith(f'.{namespace}') and hasattr(
                sys.modules[instance.__module__], public
            ):
                delattr(sys.modules[instance.__module__], public)

    def _ensure_isnamespace(self, name: str) -> None:
        """Ensure the specified name is a known namespace.

        # Returned value

        None

        # Raised exception

        A _ValueError_ exception is raised if `name` is not a known
        namespace name.
        """
        namespace, _ = self.etcd.get(
            CLUSTER_KEY_TEMPLATE.format(resource='namespaces', name=name)
        )
        if namespace is None:
            raise ValueError(status('NotFound', f'Namespace {name} not found'))

    def _ensure_isplurals(self, name: str) -> str:
        """Ensure the specified name is known.

        # Returned value

        A string, the corresponding kind.

        # Raised exceptions

        A _ValueError_ exception is raised if `name` is not a known
        plurals name.
        """
        for resources, _ in self.etcd.get_prefix(APISERVICE_PREFIX):
            definitions = json.loads(resources)['resources']
            for resource in definitions:
                if name == resource['name']:
                    return resource['kind']
        raise ValueError(status('NotFound', f'{name} not found'))

    def _get_key(self, kind: str, name: str, namespace: str) -> str:
        if namespace is None:
            return CLUSTER_KEY_TEMPLATE.format(resource=kind, name=name)
        self._ensure_isnamespace(namespace)
        return NAMESPACED_KEY_TEMPLATE.format(
            resource=kind, namespace=namespace, name=name
        )

    def _ensure_authn(self) -> str:
        """Ensure the incoming request is authenticated.

        If from localhost, assume the `'localhost'` identity.

        If from somewhere else, use the subject value in the provided
        token.

        Raises a _ValueError_ exception if the token is missing or
        invalid, with the 'Unauthorized' flag set.
        """
        if self.request.remote_addr == '127.0.0.1':
            return 'localhost'
        authz = self.request.headers.get('Authorization')
        if authz is None:
            raise ValueError(status('Unauthorized', 'No Bearer token'))
        parts = authz.split()
        if parts[0].lower() != 'bearer' or len(parts) != 2:
            raise ValueError(
                status('Unauthorized', 'Invalid Authorization header')
            )
        try:
            return jwt.decode(parts[1], self._publickey)['sub']
        except:
            raise ValueError(status('Unauthorized', 'Invalid token'))

    def _ensure_authz(self, user: str) -> None:
        """Ensure the incoming request is authorized.

        Raises a _ValueError_ exception if the request is not
        authorized, with the 'Forbidden' flag set.
        """
        if user in ('servicedesk', 'localhost') or ':' in user:
            return
        raise ValueError(status('Forbidden', 'Operation not allowed'))

    def _ensure_admission(self) -> None:
        pass

    # API

    # generic
    @entrypoint('/openapi/v2', methods=['GET'])
    def discover_openapi(self, **kwargs: Any) -> Dict[str, Any]:
        return get_openapi(self)

    @entrypoint(API_ROUTE, methods=['GET'])
    def discover_api_versions(self, **kwargs: Any) -> Dict[str, Any]:
        """Describe cluster APIs."""
        return {
            'kind': 'APIVersions',
            'versions': ['v1'],
            'serverAddressByClientCIDRs': [
                {'clientCIDR': '0.0.0.0/0', 'serverAddress': 'localhost:8080'}
            ],
        }

    @entrypoint(APIGROUP_ROUTE, methods=['GET'])
    def discover_api_groups(self, **kwargs: Any) -> Object:
        """Describe available API groups."""
        versions = {
            yaml.safe_load(r)['groupVersion']
            for r, _ in self.etcd.get_prefix(APISERVICE_PREFIX)
        }
        return {
            'kind': 'APIGroupList',
            'apiVersion': 'v1',
            'groups': [
                {
                    'name': version.split('/')[0],
                    'versions': [
                        {
                            'groupVersion': version,
                            'version': version.split('/')[1],
                        }
                    ],
                    'preferredVersion': {
                        'groupVersion': version,
                        'version': version.split('/')[1],
                    },
                }
                for version in versions
                if '/' in version
            ],
        }

    @entrypoint(APISERVICE_ROUTES, methods=['GET'])
    def discover_api_resources(
        self, group: str = '', version: str = 'v1', **kwargs: Any
    ) -> Dict[str, Any]:
        """Describe available API resources."""
        apiservice, _ = self.etcd.get(
            APISERVICE_KEY_TEMPLATE.format(group=group, version=version)
        )
        return yaml.safe_load(apiservice)

    @entrypoint(CLUSTER_ROUTES, methods=['GET'])
    def list_allnamespaces(
        self,
        kind: str,
        labelselector: str = '',
        fieldselector: str = '',
        **kwargs: Any,
    ) -> Object:
        """Return a list of matching objects in all namespaces.

        Also used to returl list of cluster-level resources.

        # Required parameters

        - kind: a non-empty string

        # Optional parameters

        - labelselector: a string (empty by default)
        - fieldselector: a string (empty by default)

        # Returned value

        A dictionary with the following entries:

        - apiVersion: a string
        - kind: a string
        - items: a possibly empty list of dictionaries
        """
        _kind = self._ensure_isplurals(kind)
        prefix = CLUSTER_PREFIX_TEMPLATE.format(resource=kind)
        return make_items(
            _kind,
            [
                json.loads(obj)
                for obj, _ in self.etcd.get_prefix(prefix)
                if _match_label_selector(json.loads(obj), labelselector)
                and _match_field_selector(json.loads(obj), fieldselector)
            ],
        )

    @entrypoint(NAMESPACED_ROUTES, methods=['GET'])
    def list_namespaced(
        self,
        kind: str,
        namespace: str = DEFAULT_NAMESPACE_NAME,
        labelselector: str = '',
        fieldselector: str = '',
        **kwargs: Any,
    ) -> Object:
        """Return a list of objects matching kind and selectors.

        # Required parameters

        - kind: a non-empty string

        # Optional parameters

        - namespace: a non-empty string (default namespace by default)
        - labelselector: a string (empty by default)
        - fieldselector: a string (empty by default)

        # Returned value

        A dictionary with the following entries:

        - apiVersion: a string
        - kind: a string
        - items: a possibly empty list of dictionaries
        """
        self._ensure_isnamespace(namespace)
        _kind = self._ensure_isplurals(kind)
        prefix = NAMESPACED_PREFIX_TEMPLATE.format(
            resource=kind, namespace=namespace
        )
        return make_items(
            _kind,
            [
                json.loads(obj)
                for obj, _ in self.etcd.get_prefix(prefix)
                if _match_label_selector(json.loads(obj), labelselector)
                and _match_field_selector(json.loads(obj), fieldselector)
            ],
        )

    @entrypoint(DIRECT_ROUTES)
    def get(
        self,
        kind: str,
        name: str,
        namespace: Optional[str] = None,
        **kwargs: Any,
    ) -> Object:
        """Return the requested object.

        # Required parameters

        - kind: a non-empty string
        - name: a non-empty string

        # Optional parameters

        - namespace: a non-empty string or None (None by default)

        # Return value

        An object (a dictionary)
        """
        self._ensure_isplurals(kind)
        obj, meta = self.etcd.get(self._get_key(kind, name, namespace))
        if obj is None:
            raise ValueError(status('NotFound', f'Object {name} not found'))
        obj = json.loads(obj)
        obj['metadata']['resourceVersion'] = str(meta.mod_revision)
        return obj

    @entrypoint(CREATE_ROUTES)
    def create(
        self,
        kind: str,
        body: Object,
        namespace: Optional[str] = None,
        **kwargs,
    ) -> Object:
        """Create a new object.

        # Required parameters

        - body: a dictionary

        # Optional parameters

        - namespace: a non-empty string or None (None by default)

        If `namespace` is specified, it overrides the metadata.namespace
        value in `body`.

        # Returned value

        The created object (a dictionary).
        """
        _kind = self._ensure_isplurals(kind)
        if body['kind'] != _kind:
            raise ValueError(
                status(
                    'Invalid', f'Mismatched kinds: {_kind} and {body["kind"]}'
                )
            )
        metadata = body['metadata']
        name = metadata['name']

        key = self._get_key(kind, name, namespace)
        item, _ = self.etcd.get(key)
        if item is not None:
            raise ValueError(
                status('AlreadyExists', f'Object {name} already exists')
            )

        if namespace is not None:
            metadata['namespace'] = namespace
        metadata['creationTimestamp'] = datetime.datetime.now().isoformat()
        metadata['uid'] = str(uuid.uuid1())
        metadata['generation'] = 1
        body['status'] = {}

        self.etcd.put(key, json.dumps(body))

        return body

    @entrypoint(DIRECT_ROUTES)
    def update(
        self,
        kind: str,
        name: str,
        body: Object,
        namespace: Optional[str] = None,
        **kwargs: Any,
    ) -> Object:
        """Replace an existing object.

        # Required parameters

        - kind: a non-empty string
        - name: a non-empty string
        - body: a dictionary

        # Optional parameters

        - namespace: a non-empty string or None (None by default)

        # Returned value

        An object (a dictionary).
        """
        self._ensure_isplurals(kind)
        key = self._get_key(kind, name, namespace)
        obj, _ = self.etcd.get(key)
        if obj is None:
            raise ValueError(status('NotFound', f'Object {name} not found'))

        obj = json.loads(obj)
        if obj['metadata']['uid'] != body['metadata']['uid']:
            raise ValueError(status('Conflict', 'uid does not match'))
        if obj['metadata']['namespace'] != body['metadata']['namespace']:
            raise ValueError(status('Conflict', 'namespace does not match'))

        self.etcd.put(key, json.dumps(body))
        return body

    @entrypoint(DIRECT_ROUTES)
    def patch(
        self,
        kind: str,
        name: str,
        body: Object,
        namespace: Optional[str] = None,
        **kwargs: Any,
    ) -> Object:
        """Patch an existing object.

        # Required parameters

        - kind: a non-empty string
        - name: a non-empty string
        - body: a dictionary

        # Optional parameters

        - namespace: a non-empty string or None (None by default)

        If `namespace` is specified, it overrides the metadata.namespace
        value in `body`.

        # Returned value

        The patched object (a dictionary).
        """
        self._ensure_isplurals(kind)
        if 'generation' in body['metadata']:
            raise ValueError(status('Invalid', 'generation field in metadata'))

        key = self._get_key(kind, name, namespace)
        item, _ = self.etcd.get(key)
        if item is None:
            raise ValueError(status('NotFound', f'Object {name} not found'))

        obj = patchdict(json.loads(item), body)
        obj['metadata']['generation'] += 1
        self.etcd.put(key, json.dumps(obj))
        return obj

    # def delete(self, ...):

    @entrypoint(STATUS_ROUTES, methods=['PATCH'])
    def patch_status(
        self,
        kind: str,
        name: str,
        body: Object,
        namespace: Optional[str] = None,
        **kwargs: Any,
    ) -> Object:
        pass

    @entrypoint(STATUS_ROUTES, methods=['UPDATE'])
    def update_status(
        self,
        kind: str,
        name: str,
        body: Object,
        namespace: Optional[str] = None,
        **kwargs: Any,
    ) -> Object:
        pass

    @entrypoint(STATUS_ROUTES, methods=['GET'])
    def get_status(
        self,
        kind: str,
        name: Object,
        namespace: Optional[str] = None,
        **kwargs: Any,
    ) -> Object:
        pass
